@@ -1,7 +1,7 @@
 import asyncio
 import json
 import logging
-from asyncio import Queue
+from asyncio import Queue, Lock
 from asyncio.futures import Future
 from logging import Logger
 from typing import Dict, Union, Any, Optional, List, Callable, Awaitable
@@ -22,7 +22,10 @@ class AsyncBaseSocketModeClient:
     app_token: str
     wss_uri: str
     auto_reconnect_enabled: bool
+    trace_enabled: bool
     closed: bool
+    connect_operation_lock: Lock
+
     message_queue: Queue
     message_listeners: List[
         Union[
@@ -58,15 +61,36 @@ class AsyncBaseSocketModeClient:
                 self.logger.error(f"Failed to retrieve WSS URL: {e}")
                 raise e
 
+    async def is_connected(self) -> bool:
+        return False
+
+    async def session_id(self) -> str:
+        return ""
+
     async def connect(self):
         raise NotImplementedError()
 
     async def disconnect(self):
         raise NotImplementedError()
 
-    async def connect_to_new_endpoint(self):
-        self.wss_uri = await self.issue_new_wss_url()
-        await self.connect()
+    async def connect_to_new_endpoint(self, force: bool = False):
+        session_id = await self.session_id()
+        try:
+            await self.connect_operation_lock.acquire()
+            if self.trace_enabled:
+                self.logger.debug(
+                    f"For reconnection, the connect_operation_lock was acquired (session: {session_id})"
+                )
+            if force or not await self.is_connected():
+                self.wss_uri = await self.issue_new_wss_url()
+                await self.connect()
+        finally:
+            if self.connect_operation_lock.locked() is True:
+                self.connect_operation_lock.release()
+                if self.trace_enabled:
+                    self.logger.debug(
+                        f"The connect_operation_lock for reconnection was released (session: {session_id})"
+                    )
 
     async def close(self):
         self.closed = True
@@ -87,16 +111,27 @@ class AsyncBaseSocketModeClient:
         await self.message_queue.put(message)
         if self.logger.level <= logging.DEBUG:
             queue_size = self.message_queue.qsize()
+            session_id = await self.session_id()
             self.logger.debug(
-                f"A new message enqueued (current queue size: {queue_size})"
+                f"A new message enqueued (current queue size: {queue_size}, session: {session_id})"
             )
 
     async def process_messages(self):
-        while not self.closed:
-            try:
-                await self.process_message()
-            except Exception as e:
-                self.logger.exception(f"Failed to process a message: {e}")
+        session_id = await self.session_id()
+        try:
+            while not self.closed:
+                try:
+                    await self.process_message()
+                except Exception as e:
+                    self.logger.exception(
+                        f"Failed to process a message: {e}, session: {session_id}"
+                    )
+        except asyncio.CancelledError:
+            if self.trace_enabled:
+                self.logger.debug(
+                    f"The running process_messages task for {session_id} is now cancelled"
+                )
+            raise
 
     async def process_message(self):
         raw_message = await self.message_queue.get()
@@ -109,21 +144,24 @@ class AsyncBaseSocketModeClient:
             )
 
     async def run_message_listeners(self, message: dict, raw_message: str) -> None:
+        session_id = await self.session_id()
         type, envelope_id = message.get("type"), message.get("envelope_id")
         if self.logger.level <= logging.DEBUG:
             self.logger.debug(
-                f"Message processing started (type: {type}, envelope_id: {envelope_id})"
+                f"Message processing started (type: {type}, envelope_id: {envelope_id}, session: {session_id})"
             )
         try:
             if message.get("type") == "disconnect":
-                await self.connect_to_new_endpoint()
+                await self.connect_to_new_endpoint(force=True)
                 return
 
             for listener in self.message_listeners:
                 try:
                     await listener(self, message, raw_message)
                 except Exception as e:
-                    self.logger.exception(f"Failed to run a message listener: {e}")
+                    self.logger.exception(
+                        f"Failed to run a message listener: {e}, session: {session_id}"
+                    )
 
             if len(self.socket_mode_request_listeners) > 0:
                 request = SocketModeRequest.from_dict(message)
@@ -133,12 +171,17 @@ class AsyncBaseSocketModeClient:
                             await listener(self, request)
                         except Exception as e:
                             self.logger.exception(
-                                f"Failed to run a request listener: {e}"
+                                f"Failed to run a request listener: {e}, session: {session_id}"
                             )
         except Exception as e:
-            self.logger.exception(f"Failed to run message listeners: {e}")
+            self.logger.exception(
+                f"Failed to run message listeners: {e}, session: {session_id}"
+            )
         finally:
             if self.logger.level <= logging.DEBUG:
                 self.logger.debug(
-                    f"Message processing completed (type: {type}, envelope_id: {envelope_id})"
+                    f"Message processing completed ("
+                    f"type: {type}, "
+                    f"envelope_id: {envelope_id}, "
+                    f"session: {session_id})"
                 )
